#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <alloca.h>
#include "nsapp.h"
#include "hw.h"
#include "board.h"
#include "sha1.h"
#include "md5.h"

board_t *boards;
int nboards;

pthread_mutex_t wu_mutex;
wu_t *wu_todo;
wu_t *wu_done;
int wu_todo_cnt, wu_todo_max;

int wu_push_done(wu_t *wu)
{
	int percent;
	if(wu_lock())
		return 0;
	wu->next = wu_done;
	wu_done = wu;
	wu_todo_cnt--;
	if(wu_todo_max)
		percent = (100*(wu_todo_max - wu_todo_cnt)) / wu_todo_max;
	else	// can't happen
		percent = 100;
	wu_unlock();
	return percent;
}

static void wu_push_todo_int(wu_t *wu)
{
	wu->next = wu_todo;
	wu_todo = wu;
}

wu_t *wu_pop_todo(void)
{
	wu_t *res = NULL;
	if(wu_lock())
		return NULL;
	if(wu_todo) {
		res = wu_todo;
		wu_todo = wu_todo->next;
		res->next = NULL;
	}
	return res;
}

void wu_push_todo(wu_t *wu)
{
	wu_push_todo_int(wu);
	wu_unlock();
}

void create_wus(task_t *task)
{
	wu_t *wu;
	int cntr[MAX_LEN], i, done;
	if(wu_lock())
		return;
	for(i=0;i<MAX_LEN;i++)
		cntr[i] = 0;

	do {
		wu = calloc(1, sizeof(wu_t));
		for(i=0; i<task->pfx_len; i++)
			wu->prefix[i] = task->prefix[i];
		for(; i<task->len-task->wu_size; i++)
			wu->prefix[i] = hamilton[cntr[i]];
		wu->task = task;

		wu_todo_max++;
		wu_todo_cnt++;
		wu_push_todo_int(wu);

		done = 1;
		for(i=task->pfx_len; i<task->len-task->wu_size; i++) {
			cntr[i]++;
			if(cntr[i]!=64) {
				done = 0;
				break;
			}
			cntr[i]=0;
		}
	} while(!done);
	wu_unlock();
}

int task_compile_map(task_t *task)
{
	unsigned *tmp;
	int i;
	if(task->hashlen == 1)
		tmp = task->hashes;
	else {
		tmp = alloca(task->nhash * sizeof(unsigned));
		for(i=0; i<task->nhash; i++)
			tmp[i] = task->hashes[i*task->hashlen];
	}
	task->map = map_compile(task->nhash, tmp);
	return !task->map;
}

void create_boards(void)
{
	char **names = hw_find_boards();
	int i;
	if(!names)
		exit(1);
	for(i=0; names[i]; i++) ;
	nboards = i;
	boards = calloc(nboards, sizeof(board_t));
	for(i=0; names[i]; i++) {
		boards[i].id = i;
		strncpy(boards[i].name, names[i], 8);
	}
	free(names);
}

void start_threads(void)
{
	int i;
	for(i=0; i<nboards; i++)
		if(pthread_create(&boards[i].thread, NULL, board_thread, &boards[i]))
			message(MSG_FAIL, "Thread failed to spawn for board %s", boards[i].name);
}

void wait_for_threads(void)
{
	int i;
	for(i=0; i<nboards; i++)
		if(boards[i].thread)
			pthread_join(boards[i].thread, NULL);
}

static FILE *msg_log;

unsigned hashes[] = {
#include "hashes.h"
};
int main(void)
{
	task_t task;

	msg_log = fopen("nsapp.log", "w");

	alg_register(&sha1_alg);
	alg_register(&md5_alg);

	task.alg = &md5_alg;
	task.len = 8;
	strcpy(task.prefix, "");
	task.pfx_len = strlen(task.prefix);
	task.hashlen = 4;
	task.nhash = sizeof(hashes)/(sizeof(unsigned)*task.hashlen);
	task.hashes = hashes;
	task.wu_size = (task.len<5)?task.len:(task.len>7)?6:(task.len-1);

	if(!task.alg)
		retval_message(1, MSG_FAIL, "Unknown algorithm specified");
	if(task_compile_map(&task))
		retval_message(1, MSG_FAIL, "Lookup map compilation failed");
	pthread_mutex_init(&wu_mutex, NULL);
	create_wus(&task);
	create_boards();
	start_threads();
	wait_for_threads();
	pthread_mutex_destroy(&wu_mutex);

	fclose(msg_log);

	return 0;
}

void message(int severity, char *warning, ...)
{
	static char *sev_str[]={"debug", "Info", "WARNING", "**ERROR**"};

	char *str;
	va_list ap;

	va_start(ap, warning);
	str=alloca(vsnprintf(NULL, 0, warning, ap)+1);
	va_end(ap);
	va_start(ap, warning);
	vsprintf(str, warning, ap);
	va_end(ap);

	if(severity > 0)
		fprintf(stderr, "%s: %s\n", sev_str[severity], str);
	fprintf(msg_log, "%s: %s\n", sev_str[severity], str);
	fflush(msg_log);
}

int wu_lock(void)
{
	if(pthread_mutex_lock(&wu_mutex))
		retval_message(1, MSG_FAIL, "Failed to lock Work Unit queue mutex");
	return 0;
}

int wu_unlock(void)
{
	if(pthread_mutex_unlock(&wu_mutex))
		retval_message(1, MSG_FAIL, "Failed to unlock Work Unit queue mutex");
	return 0;
}

unsigned char hamilton[]="\000ABCFGDELMNOJKHIXYZ_VWTUPQRSrspqtuvwfgdelmnojkhixyz8932bca145760";
